#include <lektor/common.h>
#include <lektor/cmd.h>
#include <wait.h>

static const char *executable_name = NULL;

void
cmd_set_executable_name(const char *str)
{
    if (executable_name != NULL)
        free((void *)executable_name);
    executable_name = strdup(str);
    lkt_assert(executable_name);
}

const char *
cmd_get_executable_name(void)
{
    return executable_name;
}

DESTRUCTOR_FUNCTION
___cleanup_executable_name(void)
{
    if (executable_name != NULL)
        free((void *)executable_name);
}

EXIT_FUNCTION
print_help(void)
{
    if (!executable_name)
        exit(EXIT_FAILURE);

    pid_t child_pid = fork();

    /* The child, i.e. the man page */
    if (child_pid == 0) {
        const char *const args[] = { LKT_MAN_BINARY, executable_name, NULL };
        execv(args[0], (char *const *)args);
        exit(EXIT_FAILURE);
    }

    /* The parent */
    else if (child_pid < 0) {
        LOG_FATAL("Failed to fork, can't execute " LKT_MAN_BINARY);
    }

    else {
        int wstatus;
        errno = 0;
        do {
            if (waitpid(child_pid, &wstatus, WUNTRACED | WCONTINUED) == -1)
                LOG_FATAL("Failed to wait for children: %s", strerror(errno));
        } while (!WIFEXITED(wstatus) && !WIFSIGNALED(wstatus));
        exit(EXIT_SUCCESS);
    }
}

PRIVATE_FUNCTION size_t
___get_max_options(struct cmd_env *env)
{
    struct cmd_env *it = env;
    size_t ret         = 0;

    FOR_EACH_FLAT_LIST_ITEM (it, name) {
        ret++;
    }

    return ret;
}

PRIVATE_FUNCTION bool
___find_and_set_option(struct cmd_env *env, const char *name, size_t name_len, const char *value)
{
    struct cmd_env *it = env;

    FOR_EACH_FLAT_LIST_ITEM (it, name) {
        if (STR_NMATCH(it->name, name, name_len)) {
            it->value = value;
            return true;
        }
    }

    return false;
}

void
cmd_apply_env(struct cmd_env *env, void *user)
{
    struct cmd_env *it = env;

    FOR_EACH_FLAT_LIST_ITEM (it, name) {
        it->handler(it->name, it->value, user);
    }
}

void
cmd_parse_env(struct cmd_env *env, int *argc, const char ***argv)
{
    lkt_assert((*argc) >= 0);

    size_t max_options = ___get_max_options(env);
    size_t len         = 0;
    int got            = 0;

    for (size_t i = 1; i < (size_t)(*argc) && i < max_options; ++i) {
        len = strcspn((*argv)[i], "=:");
        got += ___find_and_set_option(env, (*argv)[i], len, (*argv)[i] + len + 1);
    }

    /* Skip the gotten options from the command line. */
    *argv = &(*argv)[got + 1];
    *argc = (*argc) - (got + 1);
}

EXIT_FUNCTION
cmd_parse(struct cmd_opt *opts, int argc, const char **argv)
{
    int count            = 0, is_ok;
    struct cmd_opt *it   = opts;
    cmd_callback call[2] = { NULL, NULL };

    if (argc == 0 || *argv == NULL || !strcasecmp(argv[0], "help") ||
        !strcasecmp(argv[0], "--help") || !strcasecmp(argv[0], "-help"))
        goto help;

    /* Find the command */
    FOR_EACH_FLAT_LIST_ITEM (it, name) {
        is_ok       = (!strncasecmp(argv[0], it->name, strlen(argv[0])));
        call[is_ok] = it->call;
        count += is_ok;
    }

    /* Now search for a unique match */
    if (count > 1)
        goto not_exclusive;

    if (!call[1] || count == 0)
        goto not_found;

    /* This call should exit. */
    {
        struct cmd_args arguments = CMD_ARGS_NEXT_FROM(argc, argv);
        call[1](&arguments);
        abort();
    }

not_found:
    /* The default function */
    if (!it->name && it->call) {
        struct cmd_args arguments = CMD_ARGS_FROM(argc, argv);
        it->call(&arguments);
    }

    LOG_ERROR("COMMAND", "Command '%s' could not be found", argv[0]);
    exit(EXIT_FAILURE);

help:
    print_help();

not_exclusive:
    LOG_ERROR("COMMAND", "Failed to determine option, '%s' not exclusive", argv[0]);
    abort();
}
