#define _POSIX_C_SOURCE 200809L

#include <lektor/common.h>
#include <lektor/database.h>
#include <lektor/internal/dbmacro.h>

struct sql_field_description {
    const LKT_DATABASE_VERSION minimal_version;
    const char *name;
    const char *fallback_field;
};

struct sql_table_description {
    const char *name;
    const struct sql_field_description fields[LKT_TABLE_FIELDS_MAX];
    const LKT_DATABASE_VERSION minimal_version;
};

#define LKT_DATABASE_VERSION_ALPHA  LKT_DATABASE_VERSION_ALPHA
#define LKT_DATABASE_VERSION_MK_7_1 LKT_DATABASE_VERSION_MK_7_1
#define LKT_DATABASE_VERSION_MK_7_3 LKT_DATABASE_VERSION_MK_7_3

#define DCL_SQL_FIELD(field_name, version)                                                      \
    ((const struct sql_field_description){ .name            = field_name,                       \
                                           .minimal_version = (LKT_DATABASE_VERSION_##version), \
                                           .fallback_field  = NULL })

#define DCL_SQL_FIELD_WITH_FALLBACK(field_name, fallback_field_name, version)                   \
    ((const struct sql_field_description){ .name            = field_name,                       \
                                           .minimal_version = (LKT_DATABASE_VERSION_##version), \
                                           .fallback_field  = fallback_field_name })

#define DCL_SQL_FIELD_NULL                 \
    ((const struct sql_field_description){ \
        .name = NULL, .minimal_version = 0, .fallback_field = NULL })

#define DCL_SQL_TABLE(tbl_name, ...)                                      \
    {                                                                     \
        .name = #tbl_name, .fields = { __VA_ARGS__, DCL_SQL_FIELD_NULL }, \
        .minimal_version = LKT_DATABASE_VERSION_ALPHA,                    \
    }

#define DCL_SQL_TABLE_VERSION(tbl_name, version, ...)                     \
    {                                                                     \
        .name = #tbl_name, .fields = { __VA_ARGS__, DCL_SQL_FIELD_NULL }, \
        .minimal_version = LKT_DATABASE_VERSION_##version,                \
    }

static const struct sql_table_description sql_tables_list[] = {
    DCL_SQL_TABLE(kara,                                                                   //
                  DCL_SQL_FIELD("id", ALPHA),                                             //
                  DCL_SQL_FIELD("song_name", ALPHA),                                      //
                  DCL_SQL_FIELD("source_name", ALPHA),                                    //
                  DCL_SQL_FIELD("category", ALPHA),                                       //
                  DCL_SQL_FIELD("song_type", ALPHA),                                      //
                  DCL_SQL_FIELD("song_number", ALPHA),                                    //
                  DCL_SQL_FIELD("language", ALPHA),                                       //
                  DCL_SQL_FIELD("file_path", ALPHA),                                      //
                  DCL_SQL_FIELD("is_new", ALPHA),                                         //
                  DCL_SQL_FIELD("author_name", ALPHA),                                    //
                  DCL_SQL_FIELD("available", ALPHA),                                      //
                  DCL_SQL_FIELD("cached_mtime", MK_7_1),                                  //
                  DCL_SQL_FIELD("cached_duration", MK_7_1),                               //
                  DCL_SQL_FIELD_WITH_FALLBACK("repo_timestamp", "cached_mtime", MK_7_3)), //

    DCL_SQL_TABLE(kara_type,                     //
                  DCL_SQL_FIELD("id", ALPHA),    //
                  DCL_SQL_FIELD("name", ALPHA)), //

    DCL_SQL_TABLE(kara_category,                 //
                  DCL_SQL_FIELD("id", ALPHA),    //
                  DCL_SQL_FIELD("name", ALPHA)), //

    DCL_SQL_TABLE(language,                      //
                  DCL_SQL_FIELD("id", ALPHA),    //
                  DCL_SQL_FIELD("name", ALPHA)), //

    DCL_SQL_TABLE(playlist,                             //
                  DCL_SQL_FIELD("id", ALPHA),           //
                  DCL_SQL_FIELD("name", ALPHA),         //
                  DCL_SQL_FIELD("last_update", ALPHA)), //

    DCL_SQL_TABLE(kara_playlist,                        //
                  DCL_SQL_FIELD("kara_id", ALPHA),      //
                  DCL_SQL_FIELD("playlist_id", ALPHA)), //

    DCL_SQL_TABLE(queue,                             //
                  DCL_SQL_FIELD("position", ALPHA),  //
                  DCL_SQL_FIELD("kara_id", ALPHA),   //
                  DCL_SQL_FIELD("priority", ALPHA)), //

    DCL_SQL_TABLE(users,                             //
                  DCL_SQL_FIELD("username", ALPHA),  //
                  DCL_SQL_FIELD("password", ALPHA)), //

    DCL_SQL_TABLE(updates,                          //
                  DCL_SQL_FIELD("job", ALPHA),      //
                  DCL_SQL_FIELD("kara_id", ALPHA)), //

    DCL_SQL_TABLE(misc,                                    //
                  DCL_SQL_FIELD("id", ALPHA),              //
                  DCL_SQL_FIELD("last_update", ALPHA),     //
                  DCL_SQL_FIELD("last_end_update", ALPHA), //
                  DCL_SQL_FIELD("update_job", ALPHA),      //
                  DCL_SQL_FIELD("opened", ALPHA),          //
                  DCL_SQL_FIELD("obfuscation", ALPHA)),    //

    DCL_SQL_TABLE('stickers',                    //
                  DCL_SQL_FIELD("id", ALPHA),    //
                  DCL_SQL_FIELD("name", ALPHA)), //

    DCL_SQL_TABLE('stickers.kara',                 //
                  DCL_SQL_FIELD("id", ALPHA),      //
                  DCL_SQL_FIELD("sticker", ALPHA), //
                  DCL_SQL_FIELD("value", ALPHA)),  //

    DCL_SQL_TABLE('stickers.plt',                  //
                  DCL_SQL_FIELD("id", ALPHA),      //
                  DCL_SQL_FIELD("sticker", ALPHA), //
                  DCL_SQL_FIELD("value", ALPHA)),  //
};

static const size_t sql_tables_list_count =
    sizeof(sql_tables_list) / sizeof(struct sql_table_description);

CONSTRUCTOR_FUNCTION
___verify_table_descriptions(void)
{
    for (size_t i = 0; i < sql_tables_list_count; ++i) {
        const struct sql_table_description *table = &sql_tables_list[i];
        size_t j                                  = LKT_TABLE_FIELDS_MAX - 1;
        while (j) {
            if (table->fields[j].name == NULL)
                break;
            --j;
        }
        if (j <= 0) {
            LOG_FATAL("The description for the table '%s' has to many fields, "
                      "the maximum supported is %lu",
                      table->name, LKT_TABLE_FIELDS_MAX);
        }
    }
}

PRIVATE_FUNCTION bool
___compute_upgrade_sql_query(LKT_DATABASE_VERSION base_version,
                             const struct sql_table_description *table, char *buffer,
                             size_t buffer_length)
{
    /* The query template:
     * - 0: table name
     * - 1: all the fields, comma separated
     * - 2: same as 1
     * - 3: same as 0 */
    static const char *SQL_UPGRADE_QUERY_TEMPLATE =
        "INSERT OR REPLACE INTO " LKT_PROTECTED_DATABASE_NEW ".%s (%s)"
        "  SELECT %s FROM " LKT_PROTECTED_DATABASE ".%s;";

    /* Create the comma separated list for the table's fields */
    size_t field_list_new_length = 0;
    size_t field_list_old_length = 0;
    size_t added_fields          = 0;
    for (size_t i = 0; table->fields[i].name != NULL; ++i) {
        if (table->fields[i].minimal_version <= base_version) {
            added_fields += 1;

            /* New table field list, add the comma to the total size */
            field_list_new_length += 1;
            field_list_new_length += strlen(table->fields[i].name);

            /*
            ** Old table field list, add the comma to the total size and take
            ** into account the copy fallback fields when upgrading
            */

            if (table->fields[i].fallback_field != NULL) {
                field_list_old_length += 1;
                field_list_new_length += strlen(table->fields[i].fallback_field);
            } else {
                field_list_old_length += 1;
                field_list_old_length += strlen(table->fields[i].name);
            }
        }
    }
    if (added_fields == 0) {
        LOG_ERROR("DB-UPGRADE", "No fields where found for the database table %s", table->name);
        return false;
    }

    if ((field_list_new_length >= LKT_MAX_SQLITE_STATEMENT) ||
        (field_list_old_length >= LKT_MAX_SQLITE_STATEMENT)) {
        LOG_ERROR("DB-UPGRADE",
                  "The comma separated list for all the fields "
                  "in the table '%s' will overflow, failed to "
                  "construct the upgrade SQL query!",
                  table->name);
        return false;
    }

    char field_list_new[LKT_MAX_SQLITE_STATEMENT];
    char field_list_old[LKT_MAX_SQLITE_STATEMENT];
    memset(field_list_new, 0, sizeof(field_list_new));
    memset(field_list_old, 0, sizeof(field_list_old));

    for (size_t i = 0; table->fields[i].name; ++i) {
        if (table->fields[i].minimal_version <= base_version) {
            strcat(field_list_new, table->fields[i].name);
            strcat(field_list_new, ",");
            if (table->fields[i].fallback_field != NULL) {
                strcat(field_list_old, table->fields[i].fallback_field);
                strcat(field_list_old, ",");
            } else {
                strcat(field_list_old, table->fields[i].name);
                strcat(field_list_old, ",");
            }
        }
    }
    field_list_new[strlen(field_list_new) - 1] = '\0'; /* Crop the last "," */
    field_list_old[strlen(field_list_old) - 1] = '\0'; /* Crop the last "," */

    /* Generate the SQL query */
    memset(buffer, 0, buffer_length * sizeof(char));
    const ssize_t length = safe_snprintf(buffer, buffer_length, SQL_UPGRADE_QUERY_TEMPLATE,
                                         table->name, field_list_new, field_list_old, table->name);

    /* No overflow -> truncated query, we will have a problem... */
    return (length > 0) && ((buffer_length - 1) >= (size_t)length);
}

int
database_upgrade_scheme(va_list *db_list_ptr)
{
    lkt_db *db                   = va_arg((*db_list_ptr), lkt_db *);
    LKT_DATABASE_VERSION version = database_get_version(db);
    if (version == LKT_DATABASE_VERSION_LATEST) {
        LOG_INFO("DB-UPGRADE", "Don't upgrade the database scheme, already at the latest version");
        return true;
    }

    /* Create a new db with the new schema and attach it */
    char new_db_path[PATH_MAX]; /* New db, will be moved to the db_path */
    char bak_db_path[PATH_MAX]; /* Backup de db_path */
    char db_path[PATH_MAX];     /* The production database location */
    RETURN_UNLESS(database_config_get_text(db, "database", "db_path", new_db_path, PATH_MAX),
                  "Failed to get the path to the db file", false)
    safe_strncpy(db_path, new_db_path, PATH_MAX);
    safe_strncpy(bak_db_path, new_db_path, PATH_MAX);
    strncat(new_db_path, ".new", strlen(new_db_path) - sizeof(".new")); /* Includes '\0' */
    strncat(bak_db_path, ".bak", strlen(new_db_path) - sizeof(".bak")); /* Includes '\0' */

    if (!access(new_db_path, R_OK | W_OK)) {
        LOG_WARN("DB-UPGRADE",
                 "File '%s' already present, "
                 "delete it before continuing",
                 new_db_path);
        RETURN_IF(unlink(new_db_path), "Failed to unlink already existing file", false);
    }

    RETURN_UNLESS(database_init(new_db_path), "Failed to init new DB!", false);
    RETURN_UNLESS(database_unprotected_attach(db, LKT_PROTECTED_DATABASE_NEW, new_db_path),
                  "Failed to attach the new database disk", false);

    /* Copy the content of the old db into the new one */
    char sql_upgrade_query[LKT_MAX_SQLITE_STATEMENT];
    for (size_t i = 0; i < sql_tables_list_count; ++i) {
        const struct sql_table_description *sql_table = &sql_tables_list[i];
        if (sql_table->minimal_version > version) {
            LOG_INFO("DB-UPGRADE",
                     "Don't apply upgrade rule '%s' because the minimal database version "
                     "is %s and the old database is in version %s (field not available)",
                     sql_table->name, database_get_version_name(sql_table->minimal_version),
                     database_get_version_name(version));
            continue;
        }

        unless (___compute_upgrade_sql_query(version, sql_table, sql_upgrade_query,
                                             LKT_MAX_SQLITE_STATEMENT)) {
            LOG_ERROR("DB-UPGRADE", "Can't upgrade the database, detach the new one and exit");
            database_unprotected_detach(db, LKT_PROTECTED_DATABASE_NEW);
            return false;
        }

        LOG_INFO("DB-UPGRADE", "Copy table %s into new database for upgrade", sql_table->name);
        SQLITE_EXEC(db, sql_upgrade_query, error_in_upgrade_table);
    }

    /* Use the new DB */
    database_unprotected_detach(db, LKT_PROTECTED_DATABASE_NEW);
    database_unprotected_detach(db, LKT_PROTECTED_DATABASE);
    RETURN_IF(rename(db_path, bak_db_path), "Failed to move old db to the backup location", false);
    if (rename(new_db_path, db_path)) {
        LOG_ERROR("DB-UPGRADE", "Failed to move new db to the production location");
        FAIL_UNLESS(rename(bak_db_path, db_path),
                    "Failed to restore the production database from backup");
        return false;
    }
    FAIL_UNLESS(database_unprotected_attach(db, LKT_PROTECTED_DATABASE, db_path),
                "Failed to reattach database");
    return true;

error_in_upgrade_table:
    LOG_ERROR("DB-UPGRADE", "Failed to upgrade the database!");
    database_unprotected_detach(db, LKT_PROTECTED_DATABASE_NEW);
    return false;
}

const char *
database_get_version_name(LKT_DATABASE_VERSION version)
{
#define database_version(str, CAPS, i) \
    case LKT_DATABASE_VERSION_##CAPS: return str;
    switch (version) {
#include <lektor/internal/dbversion.def>
    case ___LKT_DATABASE_VERSION_COUNT:
        LOG_ERROR("DB", "Unknown database version %d", version);
        break;
    }
#undef database_version
    return "UNKNOWN";
}

LKT_DATABASE_VERSION
database_get_version(lkt_db *db)
{
    static const char *SQL   = "SELECT version FROM " LKT_PROTECTED_DATABASE ".misc;";
    sqlite3_stmt *stmt       = NULL;
    LKT_DATABASE_VERSION ret = LKT_DATABASE_VERSION_ALPHA;

    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_STEP_ROW(db, stmt, error);
    const char *dbversion = sqlite3_column_chars(stmt, 0);

    LOG_DEBUG("DB", "Got version %s", dbversion);

    if (NULL == dbversion)
        ret = LKT_DATABASE_VERSION_ALPHA;
#define database_version(str, CAPS, i) \
    else if (STR_MATCH(str, dbversion)) ret = LKT_DATABASE_VERSION_##CAPS;
#include <lektor/internal/dbversion.def>
#undef database_version

error:
    sqlite3_finalize(stmt);
    return ret;
}
