use crate::constants;
use amadeus_lib::actions::Action;
use amadeus_macro::*;

use super::render_action_menu;

/// Structure used to hole informations about the window to create, we use a
/// builder pattern here.
pub struct WindowBuilder<'a> {
    name: &'a str,
    dark_mode: bool,
    resizable: Option<bool>,
    default_size: Option<egui::Vec2>,
    actions: Option<Vec<Action>>,
    warning_actions: Option<Vec<Action>>,
}

impl<'a> WindowBuilder<'a> {
    pub fn new(name: &'a str, dark_mode: bool) -> Self {
        Self {
            name,
            dark_mode,
            resizable: None,
            default_size: None,
            actions: None,
            warning_actions: None,
        }
    }

    pub fn with_resizable(self, resizable: bool) -> Self {
        Self {
            name: self.name,
            dark_mode: self.dark_mode,
            resizable: Some(resizable),
            default_size: self.default_size,
            actions: self.actions,
            warning_actions: self.warning_actions,
        }
    }

    pub fn with_default_size(self, default_size: egui::Vec2) -> Self {
        Self {
            name: self.name,
            dark_mode: self.dark_mode,
            resizable: self.resizable,
            default_size: Some(default_size),
            actions: self.actions,
            warning_actions: self.warning_actions,
        }
    }

    pub fn with_actions(self, actions: Vec<Action>) -> Self {
        Self {
            name: self.name,
            dark_mode: self.dark_mode,
            resizable: self.resizable,
            default_size: self.default_size,
            actions: Some(actions),
            warning_actions: self.warning_actions,
        }
    }

    pub fn with_warning_actions(self, actions: Vec<Action>) -> Self {
        Self {
            name: self.name,
            dark_mode: self.dark_mode,
            resizable: self.resizable,
            default_size: self.default_size,
            actions: self.actions,
            warning_actions: Some(actions),
        }
    }

    pub fn show<R>(
        self,
        flag: &mut bool,
        ctx: &egui::Context,
        add_contents: impl FnOnce(&mut egui::Ui) -> R,
    ) -> Option<Vec<Action>> {
        let mut flag_copy = *flag;
        let mut temp_actions = Vec::<Action>::new();

        // Force the mut borrow of the flag to be dropped
        {
            let win = egui::Window::new(self.name).open(flag).title_bar(false);
            let (win, resizable) = match self.resizable {
                Some(resizable) => (win.resizable(resizable), resizable),
                None => (win, false),
            };
            let win = match self.default_size {
                None => win,
                Some(sizes) => {
                    either!(resizable => win.default_size(sizes).fixed_size(sizes); win.default_size(sizes))
                }
            };
            win.show(ctx, |ui| {
                ui.horizontal(|ui| {
                    ui.style_mut().override_text_style = Some(super::utils::font::heading3());
                    ui.add(egui::Button::new("❌"))
                        .on_hover_text("Close the window")
                        .clicked()
                        .then(|| flag_copy = false);

                    self.actions
                        .map(|actions| render_action_menu(ui, &actions, "▶", &mut temp_actions));
                    self.warning_actions
                        .map(|actions| render_action_menu(ui, &actions, "⚠", &mut temp_actions));

                    ui.colored_label(constants::get_text_color(self.dark_mode), self.name);
                });
                ui.add_space(constants::PADDING / 2.);
                add_contents(ui)
            });
        }

        // See if we need to close the window
        *flag = flag_copy;

        // Return the actions if needed
        if temp_actions.is_empty() {
            None
        } else {
            Some(temp_actions)
        }
    }
}
